#include <cstdlib>
#include <iomanip>
#include <iostream>
#include "math.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <cuda_runtime_api.h>


#include "eigen.h"

//  ,    ,  - 
//#define TEST

int main()
{
    // .
#ifdef TEST
    const int n = BLOCK_SIZE;
#else
    const int n = BLOCK_SIZE * 4;
#endif
    //  
    const int repeats = 100;

    float * a, * x, * y, * a_gpu, * x_gpu, * y_gpu, * x_verify, * y_host;

    // a   n x n
    a = new float[n * n];
    cudaMalloc((void **) &a_gpu, n * n * sizeof(float));

    // x, y    n
    x = new float[n];
    cudaMalloc((void **) &x_gpu, n * sizeof(float));
    x_verify = new float[n];
    y = new float[n];
    cudaMalloc((void **) &y_gpu, n * sizeof(float));
    y_host = new float[n];

#ifdef TEST
    //   diag(1, 2, ..., n):
    for (int i = 0; i < n; ++i)
        for (int j = 0; j < n; ++j)
            if (i != j)
                a[i * n + j] = 0;
            else
                a[i * n + j] = i;
#else
    //   y  .
    for (int i = 0; i < n; ++i)
        for (int j = 0; j < i; ++j)
        {
            float number = float(rand()) / RAND_MAX;
            a[i * n + j] = number;
            a[j * n + i] = number;
        }
#endif

    //     
    //   1.
    for (int i = 0; i < n; ++i)
        x[i] = 1.0f;

    //     GPU.
    cudaMemcpy(a_gpu, a, n * n * sizeof(float), cudaMemcpyHostToDevice);
    cudaMemcpy(x_gpu, x, n * sizeof(float), cudaMemcpyHostToDevice);

    //       .
    LARGE_INTEGER before, after, freq;
    QueryPerformanceFrequency(&freq);
    std::cout << std::fixed << std::setprecision(3);

    QueryPerformanceCounter(&before);
    for (int i = 0; i < repeats; ++i)
        eigen(n, a, x, y);
    QueryPerformanceCounter(&after);
    double cpu_time = (after.QuadPart - before.QuadPart) / double(freq.QuadPart);
    std::cout << "The CPU version took " << cpu_time << " seconds.\n";

    QueryPerformanceCounter(&before);
    for (int i = 0; i < repeats; ++i)
        eigen_gpu(n, a_gpu, x_gpu, y_gpu, y_host);
    // ,     .
    cudaThreadSynchronize();
    QueryPerformanceCounter(&after);
    double gpu_time = (after.QuadPart - before.QuadPart) / double(freq.QuadPart);

    std::cout << "The GPU version took " << gpu_time << " seconds.\n";

    std::cout << "Speedup: " << cpu_time / gpu_time << " times.\n";

#ifdef TEST
    //      .
    float lambda = fabs(y[0]);
    for (int i = 0; i < n; ++i)
        if (fabs(y[i]) > lambda)
            lambda = fabs(y[i]);
    std::cout << "Host result: max eigenvalue = " << lambda << "\n";

    cudaMemcpy(y_host, y_gpu, n * sizeof(float), cudaMemcpyDeviceToHost);
    lambda = fabs(y_host[0]);
    for (int i = 0; i < n; ++i)
        if (fabs(y_host[i]) > lambda)
            lambda = fabs(y_host[i]);
    std::cout << "Device result: max eigenvalue = " << lambda << "\n";
#endif

    //        .
    cudaMemcpy(x_verify, x_gpu, n * sizeof(float), cudaMemcpyDeviceToHost);

    //     y  y_verify.
    float difference = 0;
    for (int i = 0; i < n; ++i)
        difference += (x[i] - x_verify[i]) * (x[i] - x_verify[i]);

    if (difference < 1e-5f)
        std::cout << "Test passed.\n";
    else
        std::cout << "Test failed (diff = " << difference << ").\n";

    //  
    delete [] a;
    cudaFree(a_gpu);
    delete [] x;
    cudaFree(x_gpu);
    delete [] x_verify;
    delete [] y;
    cudaFree(y_gpu);
    delete [] y_host;
}
