#include "eigen.h"

//    y = Ax.
__global__ void mvmult_kernel(int n, const float * a, float * x, float * y)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    if (i >= n)
        return;

    float sum = 0.0f;
    for (int j = 0; j < n; ++j)
        sum += a[i * n + j] * x[j];
    y[i] = sum;
}


//    x = y/c,  - .
__global__ void norm_kernel(int n, float * x, float * y, float c)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    if (i < n)
        x[i] = y[i] / c;
}


void eigen_gpu(int n, const float * a, float * x, float * y, float * y_host)
{
    const int threads_per_block = 256;
    int num_blocks = (n + threads_per_block - 1) / threads_per_block;
    mvmult_kernel<<<num_blocks, threads_per_block>>>(n, a, x, y);

    cudaMemcpy(y_host, y, n * sizeof(float), cudaMemcpyDeviceToHost);
    //       y
    float yMax = fabs(y_host[0]);
    for (int i = 0; i < n; ++i)
        if (fabs(y_host[i]) > yMax)
            yMax = fabs(y_host[i]);

    norm_kernel<<<num_blocks, threads_per_block>>>(n, x, y, yMax);
}
